/*	$AoiMoe: mshelpserv/mshelpserv.cpp,v 1.3 2003/09/30 20:05:54 tshiozak Exp $	*/

/*-
 * Copyright (c)2003 Takuya SHIOZAKI <tshiozak@bsdclub.org>,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <winsock2.h>
#include <windows.h>
#include <stdio.h>
#include <stdarg.h>

#include <atlbase.h>
#include <atlctl.h>

#include <memory>
#include <vector>
#include <list>
#include <algorithm>
#include <functional>

using namespace std;

/// debug print.
#define DPR(a) dprintf a
void
dprintf(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
}

/*
 * literals.
 */
#define LN "\r\n"
#define WS " \t"
#define SCHEME_HTTP_STR "http://"
#define SCHEME_HTTP_STR_LEN (sizeof(SCHEME_HTTP_STR)-1)

/// non-copyable attribute.
class CNonCopyable
{
	CNonCopyable(CNonCopyable const &);
	void operator = (CNonCopyable const &);
public:
	CNonCopyable() { }
};


/// replace v into n.
template <typename T_>
T_ replace(T_ &v, T_ n)
{
	swap(v, n);
	return n;
}

/// disposer.
template <typename T_>
class CSimpleDisposer
{
public:
	static const T_ Null;
	static dispose(T_ ptr)
	{
		delete ptr;
	}
	void operator () (T_ ptr) const
	{
		this->dispose(ptr);
	}
};
template <typename T_>
const T_ CSimpleDisposer<T_>::Null = 0;

/// socket closer for scoped ptr.
class CSocketCloser
{
public:
	static const SOCKET Null;
	static void dispose(SOCKET s)
	{
		if (s!=Null)
			closesocket(s);
	}
};
const SOCKET CSocketCloser::Null = INVALID_SOCKET;

/// handle closer for scoped ptr.
class CHandleCloser
{
public:
	static const HANDLE Null;
	static void dispose(HANDLE h)
	{
		if (h!=Null)
			CloseHandle(h);
	}
};
const HANDLE CHandleCloser::Null = NULL;

/// scoped ptr.
template <typename T_, class D_ = CSimpleDisposer<T_> >
class CScopedPtr : CNonCopyable
{
	T_ m_ptr_;
public:
	CScopedPtr() : m_ptr_(D_::Null) { }
	CScopedPtr(T_ ptr) : m_ptr_(ptr) { }
	~CScopedPtr() { D_::dispose(m_ptr_); }
	T_ release()
	{
		return replace(m_ptr_, D_::Null);
	}
	void reset(T_ ptr = D_::Null)
	{
		swap(m_ptr_, ptr);
		D_::dispose(ptr);
	}
	T_ own() const
	{
		return m_ptr_;
	}
	operator bool ()
	{
		return m_ptr_ != D_::Null;
	}
	T_ operator * ()
	{
		return m_ptr_;
	}
};

typedef CScopedPtr<SOCKET, CSocketCloser> CScopedSocket;
typedef CScopedPtr<HANDLE, CHandleCloser> CScopedHandle;


/// wait slot.
class CWaitSlot
{
public:
	virtual ~CWaitSlot() = 0;
	virtual void operator () () = 0;
};
CWaitSlot::~CWaitSlot() { }

/// wait slot delegate.
template <class T_>
class CWaitSlotDelegate : public CWaitSlot
{
	typedef void (T_::*MFuncPtr)();
	T_ &m_object_;
	MFuncPtr m_func_;
public:
	CWaitSlotDelegate(T_ &obj, MFuncPtr func)
		: m_object_(obj), m_func_(func)
	{
	}
	~CWaitSlotDelegate()
	{
	}
	void operator () ()
	{
		(m_object_.*m_func_)();
	}
};

/// message slot.
class CMessageSlot
{
public:
	virtual ~CMessageSlot() = 0;
	virtual void operator () (MSG &) = 0;
};
CMessageSlot::~CMessageSlot() { }

/// message slot delegate.
template <class T_>
class CMessageSlotDelegate : public CMessageSlot
{
	typedef void (T_::*MFuncPtr)(MSG &);
	T_ &m_object_;
	MFuncPtr m_func_;
public:
	CMessageSlotDelegate(T_ &obj, MFuncPtr func)
		: m_object_(obj), m_func_(func)
	{
	}
	~CMessageSlotDelegate()
	{
	}
	void operator () (MSG &msg)
	{
		(m_object_.*m_func_)(msg);
	}
};

/// reply body slot.
class CReplyBodySlot
{
public:
	virtual ~CReplyBodySlot() = 0;
	virtual void operator () (const BYTE *, DWORD) = 0;
};
CReplyBodySlot::~CReplyBodySlot() { }
/// reply body slot delegate.
template <class T_>
class CReplyBodySlotDelegate : public CReplyBodySlot
{
	typedef void (T_::*MFuncPtr)(const BYTE *, DWORD);
	T_ &m_object_;
	MFuncPtr m_func_;
public:
	CReplyBodySlotDelegate(T_ &obj, MFuncPtr func)
		: m_object_(obj), m_func_(func)
	{
	}
	~CReplyBodySlotDelegate()
	{
	}
	void operator () (const BYTE *buf, DWORD size)
	{
		(m_object_.*m_func_)(buf, size);
	}
};

/// wait for object.
class CWaitFor : CNonCopyable
{
	vector<HANDLE> m_handles_;
	vector<CWaitSlot *> m_slots_;
	CMessageSlot *m_message_;
public:
	CWaitFor() : m_message_(NULL) { }
	~CWaitFor() { }
	void add(HANDLE h, CWaitSlot &ws)
	{
		m_handles_.push_back(h);
		m_slots_.push_back(&ws);
	}
	void remove(HANDLE h)
	{
		vector<HANDLE>::iterator i;
		int idx;

		i = find(m_handles_.begin(), m_handles_.end(), h);
		if (i != m_handles_.end()) {
			idx = i-m_handles_.begin();
			m_slots_.erase(m_slots_.begin()+idx);
			m_handles_.erase(i);
		}
	}
	void set(CMessageSlot &ms)
	{
		m_message_ = &ms;
	}
	void remove()
	{
		{
			vector<HANDLE> tmp;
			m_handles_.swap(tmp);
		}
		{
			vector<CWaitSlot *> tmp;
			m_slots_.swap(tmp);
		}
		m_message_ = NULL;
	}
	void compaction()
	{
		{
			vector<HANDLE> tmp(m_handles_);
			m_handles_.swap(tmp);
		}
		{
			vector<CWaitSlot *> tmp(m_slots_);
			m_slots_.swap(tmp);
		}
	}
	DWORD operator () (DWORD timo = INFINITE, DWORD mask = QS_ALLEVENTS)
	{
		DWORD rv, count=m_handles_.size();

		rv = MsgWaitForMultipleObjectsEx(count,
						 &m_handles_[0],
						 INFINITE,
						 mask,
						 MWMO_INPUTAVAILABLE|
						 MWMO_ALERTABLE);
		if (rv >= WAIT_OBJECT_0 && rv < WAIT_OBJECT_0+count)
			(*m_slots_[rv-WAIT_OBJECT_0])();
		else if (rv==WAIT_OBJECT_0+count || rv==WAIT_IO_COMPLETION) {
			MSG msg;

			while (PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
				if (m_message_)
					(*m_message_)(msg);
		} else
			return rv;

		return 0;
	}
};

/// downloader.
class CDownloader : CNonCopyable
{
	typedef CComSingleThreadModel _ThreadModel; // XXX
	class CReplyBodySlot &m_reply_body_;
public:
	CDownloader(CReplyBodySlot &reply) : m_reply_body_(reply) { }
	void OnData(CBindStatusCallback<CDownloader> *bsc,
		    BYTE *buf, DWORD size) throw()
	{
		m_reply_body_(buf, size);
	}
	HRESULT download(const char *url) throw()
	{
		CComBSTR bstrURL = url;
		return CBindStatusCallback<CDownloader>::Download(this,
								  OnData,
								  bstrURL);
	}
};


/// predicate for span.
class CSpanner : public unary_function<char, bool>
{
	string m_delim_;
public:
	CSpanner(const string &delim)
		: m_delim_(delim) { }
	bool operator () (char ch) const
	{
		return m_delim_.find(ch) < m_delim_.size();
	}
};

/// reply.
class CHTTPReply : CNonCopyable
{
	SOCKET m_socket_;
	string m_reply_buffer_;
	string::iterator m_reply_current_;
	int m_reply_remain_;
	void init_()
	{
		m_reply_current_ = m_reply_buffer_.begin();
		m_reply_remain_ = m_reply_buffer_.size();
	}
public:
	CHTTPReply() { }
	void setup_reply(SOCKET s, const string &rep)
	{
		m_socket_ = s;

		// create OK reply.
		char buf[100];

		m_reply_buffer_ =
		    "HTTP/1.0 200 OK\r\n"
		    "Content-Type: text/html\r\n";
		sprintf(buf, "Content-Length: %u\r\n\r\n", rep.size());
		m_reply_buffer_ += buf;
		m_reply_buffer_ += rep;

		this->init_();
	}
	void setup_error(SOCKET s, int err, const string &mes)
	{
		m_socket_ = s;

		// create error reply.
		char buf[100];

		sprintf(buf, "%03u ", err);
		m_reply_buffer_ += "HTTP/1.0 ";
		m_reply_buffer_ += buf;
		m_reply_buffer_ += mes; m_reply_buffer_ += "\r\n";
		m_reply_buffer_ += "Content-Type: text/plain\r\n\r\n";
		m_reply_buffer_ += buf;
		m_reply_buffer_ += mes; m_reply_buffer_ += "\r\n";

		this->init_();
	}
	void on_write_()
	{
		int rv;

		rv = send(m_socket_, &*m_reply_current_, m_reply_remain_, 0);
		if (rv == SOCKET_ERROR)
			throw WSAGetLastError();
		m_reply_current_ += rv;
		m_reply_remain_ -= rv;

		if (m_reply_remain_ == 0) {
			DPR(("reply done.\n"));
			throw 0; // XXX
		}
	}
};

/// process single connection.
class CHTTPProcessor : CNonCopyable
{
	CWaitFor &m_waitfor_;
	CScopedSocket m_socket_;
	CScopedHandle m_socket_event_;
	string m_read_buffer_, m_write_buffer_, m_url_;
	bool m_first_line_;
	CWaitSlotDelegate<CHTTPProcessor> m_event_delegate_;
	CReplyBodySlotDelegate<CHTTPProcessor> m_reply_body_delegate_;
	CDownloader m_downloader_;
	CHTTPReply m_reply_;
	void on_method_()
	{
//		DPR(("OnMethod_: %.*s\n",
//		     (int)m_read_buffer_.size(), &m_read_buffer_[0]));

		string::iterator i, j;

		i = find_if(m_read_buffer_.begin(), m_read_buffer_.end(),
			    CSpanner(WS));
		if (i == m_read_buffer_.end() ||
		    i-m_read_buffer_.begin() != 3 ||
		    toupper(m_read_buffer_[0]) != 'G' ||
		    toupper(m_read_buffer_[1]) != 'E' ||
		    toupper(m_read_buffer_[2]) != 'T') {
			// malformed method.
			DPR(("malformed method - %.*s\n",
			     (int)m_read_buffer_.size(), &m_read_buffer_[0]));
			this->setup_error_(400, "Bad Request");
			return;
		}
		// skip WS just after "GET" and cut off URL.
		i = find_if(i, m_read_buffer_.end(), not1(CSpanner(WS)));
		j = find_if(i, m_read_buffer_.end(), CSpanner(WS));
		if (i == m_read_buffer_.end() ||
		    j == m_read_buffer_.end()) {
			// malformed GET.
			DPR(("malformed GET - %.*s\n",
			     (int)m_read_buffer_.size(), &m_read_buffer_[0]));
			this->setup_error_(400, "Bad Request");
			return;
		}
		if (*i != '/') {
			if (j-i < SCHEME_HTTP_STR_LEN ||
			    string(i, i+SCHEME_HTTP_STR_LEN).compare(
				    SCHEME_HTTP_STR) != 0) {
				// malformed URL.
				DPR(("malformed URL - %.*s\n",
				     (int)m_read_buffer_.size(),
				     &m_read_buffer_[0]));
				this->setup_error_(400, "Bad Request");
				return;
			}
			i += SCHEME_HTTP_STR_LEN;
			// skip host part.
			i = find(i, j, '/');
			if (i == j) {
				// illegal URL --- file part is mandatory.
				DPR(("illegal URL - %.*s.\n",
				     (int)m_read_buffer_.size(),
				     &m_read_buffer_[0]));
				this->setup_error_(404, "Not Found");
				return;
			}
		}
		m_url_  = "ms-help://";
		m_url_ += string(i+1, j);

		m_downloader_.download(m_url_.c_str());
	}
	void on_socket_event_()
	{
		int rv;
		WSANETWORKEVENTS ev;

		rv = WSAEnumNetworkEvents(m_socket_.own(),
					  m_socket_event_.own(),
					  &ev);
		if (rv)
			throw WSAGetLastError();
		if ((ev.lNetworkEvents & FD_READ)) {
			// check error.
			if (ev.iErrorCode[FD_READ_BIT])
				throw ev.iErrorCode[FD_READ_BIT];

			this->on_read_();
		}
		if ((ev.lNetworkEvents & FD_WRITE)) {
			// check error.
			if (ev.iErrorCode[FD_WRITE_BIT])
				throw ev.iErrorCode[FD_WRITE_BIT];

			m_reply_.on_write_();
		}
		if ((ev.lNetworkEvents & FD_CLOSE)) {
			// close.
			DPR(("connection closed.\n"));
			throw 0;
		}
	}
	void on_read_()
	{
		int rv;
		u_long len;
		string::iterator i;
		int curpos = 0;

		rv = ioctlsocket(m_socket_.own(), FIONREAD, &len);
		if (rv)
			throw WSAGetLastError();

		if (m_first_line_)
			// append.
			curpos = m_read_buffer_.size();

		m_read_buffer_.resize(curpos+len);
		i = m_read_buffer_.begin() + curpos;
		rv = recv(m_socket_.own(), &*i, len, 0);
		if (rv == SOCKET_ERROR)
			throw WSAGetLastError();

		if (m_first_line_) {
			size_t lnpos;
			i = find_if(i, m_read_buffer_.end(),
				    CSpanner("\r\n"));
			if (i != m_read_buffer_.end()) {
				// found end of the line.
				m_read_buffer_.erase(
					i, m_read_buffer_.end());
				this->on_method_();
				m_first_line_ = false;
			}
		} // else discard data.
	}
	void setup_write_()
	{
		int rv;

		rv = WSAEventSelect(m_socket_.own(), m_socket_event_.own(),
				    FD_READ|FD_WRITE|FD_CLOSE);
		if (rv)
			delete this;
	}
	void setup_reply_()
	{
		m_reply_.setup_reply(m_socket_.own(), m_write_buffer_);
		this->setup_write_();
	}
	void setup_error_(int err, const string &mes)
	{
		m_reply_.setup_error(m_socket_.own(), err, mes);
		this->setup_write_();
	}
	void closedown_()
	{
		m_waitfor_.remove(m_socket_event_.own());
		m_socket_.reset();
		m_socket_event_.reset();
	}
	void OnSocketEvent_()
	{
		try {
			this->on_socket_event_();
		}
		catch (int rv) {
			delete this;
			if (rv)
				throw;
		}
	}
	void OnReplyBody_(const BYTE *buf, DWORD sz) throw()
	{
//		DPR(("OnReplyBody_: %p, %u\n", buf, sz));
		try {
			if (sz == 0) {
				// end of file.
				this->setup_reply_();
			} else if (buf == NULL) {
				// error.
				this->setup_error_(404, "Not Found");
			} else {
				// append to buffer.
				m_write_buffer_.append(
					reinterpret_cast<const char *>(buf),
					sz);
			}
		}
		catch (...) {
			//
		}
	}
public:
	CHTTPProcessor(CWaitFor &wf, SOCKET listener)
		: m_waitfor_(wf), m_first_line_(true),
		  m_event_delegate_(*this, &CHTTPProcessor::OnSocketEvent_),
		  m_reply_body_delegate_(*this, &CHTTPProcessor::OnReplyBody_),
		  m_downloader_(m_reply_body_delegate_)
	{
		// accept
		struct sockaddr_in sin;
		int addrlen = sizeof(sin);
		CScopedSocket as =
		    accept(listener,
			   reinterpret_cast<struct sockaddr *>(&sin),
			   &addrlen);
		if (!as)
			throw WSAGetLastError();

		DPR(("accept: peer = %s\n", inet_ntoa(sin.sin_addr)));

		// bind to event.
		CScopedHandle ev = WSACreateEvent();
		if (!ev)
			throw WSAGetLastError();
		if (WSAEventSelect(as.own(), ev.own(), FD_READ|FD_CLOSE))
			throw WSAGetLastError();

		m_waitfor_.add(ev.own(), m_event_delegate_);

		m_socket_.reset(as.release());
		m_socket_event_.reset(ev.release());
	}
	~CHTTPProcessor()
	{
		this->closedown_();
	}
};

/// server class.
class CHTTPServer : CNonCopyable
{
	CWaitFor &m_waitfor_;
	CScopedSocket m_listener_;
	CScopedHandle m_listener_event_;
	CWaitSlotDelegate<CHTTPServer> m_event_delegate_;
	void OnAccept_()
	{
		int rv;
		WSANETWORKEVENTS ev;

		rv = WSAEnumNetworkEvents(m_listener_.own(),
					  m_listener_event_.own(),
					  &ev);
		if (rv)
			throw WSAGetLastError();
		if ((ev.lNetworkEvents & FD_ACCEPT)) {
			if (ev.iErrorCode[FD_ACCEPT_BIT])
				throw ev.iErrorCode[FD_ACCEPT_BIT];
			try {
				CScopedPtr<CHTTPProcessor *> pr =
				    new CHTTPProcessor(m_waitfor_,
						       m_listener_.own());
				pr.release();
			}
			catch (...) {
				throw;
			}
		}
	}
public:
	CHTTPServer(CWaitFor &wf)
		: m_waitfor_(wf),
		  m_event_delegate_(*this, &CHTTPServer::OnAccept_)
	{
	}
	void setup_listener(unsigned short port, int backlog = 5)
	{
		int rv;

		DPR(("create listener..."));
		// create socket.
		CScopedSocket sock = socket(PF_INET, SOCK_STREAM, 0);
		if (!sock)
			throw WSAGetLastError();

		// set reuse address.
		{
			BOOL val = TRUE;
			rv = setsockopt(sock.own(), SOL_SOCKET, SO_REUSEADDR,
					reinterpret_cast<char *>(&val),
					sizeof(val));
			if (rv)
				throw WSAGetLastError();
		}
		// bind address.
		{
			struct sockaddr_in sin;

			sin.sin_family = AF_INET;
			sin.sin_port = htons(port);
			sin.sin_addr.s_addr = htonl(INADDR_ANY);
			rv = bind(sock.own(),
				  reinterpret_cast<struct sockaddr *>(&sin),
				  sizeof(sin));
			if (rv)
				throw WSAGetLastError();
		}
		// listen.
		rv = listen(sock.own(), backlog);
		if (rv)
			throw WSAGetLastError();

		// bind to event.
		CScopedHandle ev = WSACreateEvent();
		if (!ev)
			throw WSAGetLastError();
		if (WSAEventSelect(sock.own(), ev.own(), FD_ACCEPT))
			throw WSAGetLastError();

		m_waitfor_.add(ev.own(), m_event_delegate_);

		m_listener_.reset(sock.release());
		m_listener_event_.reset(ev.release());
		DPR(("done.\n"));
	}
};


/// exe module class.
class CMSHelpServ : public CAtlExeModuleT<CMSHelpServ>
{
	typedef CAtlExeModuleT<CMSHelpServ> Base_;
	bool m_done_;
	unsigned short m_port_;
	CWaitFor m_waitfor_;
	CHTTPServer m_server_;
	CMessageSlotDelegate<CMSHelpServ> m_messageDelegate_;
public:
	CMSHelpServ()
		: m_done_(false), m_port_(80), m_server_(m_waitfor_),
		  m_messageDelegate_(*this, &CMSHelpServ::OnMessage)
	{
	}
	void OnMessage(MSG &msg) throw()
	{
		if (msg.hwnd) {
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
		if (msg.message == WM_QUIT)
			m_done_ = true;
	}
	HRESULT PreMessageLoop(int) throw()
	{
		WSADATA wsad;

		WSAStartup(MAKEWORD(2, 2), &wsad);

		m_waitfor_.set(m_messageDelegate_);
		m_server_.setup_listener(m_port_);

		return S_OK;
	}
	void RunMessageLoop() throw()
	{
		DWORD rv;

		do {
			rv = m_waitfor_();
			if (m_done_)
				break;
		} while (rv==0);
	}
	static int start(unsigned short port = 80)
	{
		s_module_.m_port_ = port;
		return s_module_.WinMain(0);
	}
	static CMSHelpServ s_module_;
};
CMSHelpServ CMSHelpServ::s_module_;

class CUsage
{
	string m_mes_;
public:
	CUsage(const string &mes) : m_mes_(mes) { }
	const string &message() const { return m_mes_; }
};

int
main(int argc, char **argv)
{
	int rv;

	try {
		unsigned short opt_p = 80;
		char *opt;
		argv++;
		while (*argv && **argv == '-') {
			opt = *argv++;
			if (!strcmp(opt, "-p")) {
				// port.
				if (!*argv)
					throw CUsage("port number lacked.");
				opt_p = atoi(*argv++);
			} else if (!strcmp(opt, "--")) {
				break;
			} else {
				throw CUsage("illegal option.");
			}
		}
		rv = CMSHelpServ::start(opt_p);
	}
	catch (CUsage &cu) {
		// TBD
		DPR(("error: %s\n", cu.message().c_str()));
	}
	catch (int ex) {
		// TBD
		DPR(("error: %d\n", ex));
		rv = ex;
	}
	return rv;
}
